require 'base64'
require 'socket'
require 'ed25519'
require 'json'
require 'time'
require 'openssl'
require 'uri'
require 'inifile'
require 'concurrent'
require 'fluent/plugin/output'
require 'dbus'

module Fluent::Plugin
  class DatabusOutput < Output
    Fluent::Plugin.register_output('databus', self)
    desc 'The DataBus endpoint, e.g. https://dbusapi.sigma-dev.cloudera.com'
    config_param :endpoint, :string
    desc 'Environment region for databus, e.g. us-west-1'
    config_param :region, :string, :default => 'us-west-1'
    desc 'The ini file containing DataBus API credentials. Required properties are \'databus_access_key_id\', \'databus_access_secret_key\', and \'databus_access_secret_key_algo\'.'
    config_param :credential_file, :string
    desc 'The name of section used in credentials file. Default: dbus.'
    config_param :credential_profile_name, :string, :default => 'dbus'
    desc 'The interval in second to reload credentials file. Disable reloading by setting to 0 or less. Default: 60'
    config_param :credential_file_reload_interval, :integer, :default => 60
    desc 'The DataBus steam name.'
    config_param :stream_name, :string
    desc 'The DataBus partition key.'
    config_param :partition_key, :string
    desc 'The DataBus headers appended to every requests. Default: none.'
    config_param :headers, :hash, :default => {}
    desc 'Field name that contains additional headers, only works if event_message_field is used.'
    config_param :extra_headers_field, :string, :default => nil
    desc 'Enable debug log. Default: false.'
    config_param :debug, :bool, :default => false
    desc 'Enable debug logging of record metadata (record_id, state, headers) for every transfer. Default: false.'
    config_param :debug_record_metadata, :bool, :default => false
    desc 'Enable debug logging of the message that will be signed for the request. Default: false.'
    config_param :debug_request_signer, :bool, :default => false
    desc 'Enable ssl verification. Default: true.'
    config_param :verify_ssl, :bool, :default => true
    desc 'Customize SSL CA certificate file (search system ssl certificate by default).'
    config_param :ssl_ca_cert, :string, :default => nil
    desc 'The message field to extract from fluentd event as DataBus payload. If not set the whole event is transformed to json string as single message.'
    config_param :event_message_field, :string, :default => nil
    desc 'Proxy URL for DataBus endpoint'
    config_param :proxy_url, :string, :default => nil
    desc 'Proxy username for DataBus proxy endpoint (if proxy_url is filled)'
    config_param :proxy_username, :string, :default => nil
    desc 'Proxy password for DataBus proxy username (if proxy_url is filled)'
    config_param :proxy_password, :string, :default => nil, :secret => true
    desc 'List of urls to bypass proxy (if proxy_url is filled)'
    config_param :no_proxy, :string, :default => nil
    desc 'The time limit for HTTP request in seconds. Default: 120'
    config_param :http_timeout_seconds, :integer, :default => 120

    RSA_ALGORITHM = 'RSA'
    RSA_AUTH_METHOD = 'rsav1'
    ED25519_ALGORITHM = 'ED25519'
    ED25519_AUTH_METHOD = 'ed25519v1'

    DBUS_HOST_HEADER_KEY_NAME = 'key'
    DBUS_TAG_HEADER_KEY_NAME = 'tag'

    INLINE_RECORD_THRESHOLD = 600 * 1024

    # For testing
    attr_writer :api_client_class, :now
    attr_reader :api_client
    private :api_client_class=, :api_client, :now=

    def initialize
      super
      @api_client_class = Dbus::ApiClient
      @now = lambda {Time.now.utc.rfc822}
    end

    def configure(conf)
      super
      init_config_attr
      setup_cred_loader
      config_api
    end

    def init_config_attr
      @hostname = Socket.gethostname
      [DBUS_HOST_HEADER_KEY_NAME, DBUS_TAG_HEADER_KEY_NAME].each do |header|
        if @headers.key?(header)
          log.warn("Header #{header} is reserved by out_databus and will be overridden by default value.")
          @headers.delete(header)
        end
      end
      @proxy_config = {}
      unless @proxy_url.nil? || @proxy_url.empty?
        log.info "proxy config is filled. URL: #{@proxy_url}"
        @proxy_config[:url] = @proxy_url
        unless @proxy_username.nil? || @proxy_username.empty?
          log.info "proxy user is filled."
          @proxy_config[:username] = @proxy_username
        end
        unless @proxy_password.nil?
          log.info "proxy password is filled."
          @proxy_config[:password] = @proxy_password
        end
        if host_in_no_proxy(@no_proxy, @endpoint)
          @proxy_config[:url] = nil
        end
      end
    end

    def setup_cred_loader
      @load_cred_lock = Concurrent::ReadWriteLock.new
      load_cred

      if @credential_file_reload_interval > 0
        log.info("Start reloading credentials every #{@credential_file_reload_interval} seconds.")
        @load_cred_task = Concurrent::TimerTask.new(
            execution_interval: @credential_file_reload_interval) {
          begin
            load_cred
          rescue Exception => e
            log.warn("#{e.message}, continue with previous credentials.")
          end
        }
        @load_cred_task.execute
      end
    end

    def load_cred
      cred_file = IniFile.load(@credential_file) or
          raise Fluent::ConfigError, "Failed to load credential file #{@credential_file}"
      section = cred_file[@credential_profile_name] or
          raise Fluent::ConfigError, "Section #{@credential_profile_name} not found."
      new_access_key = section['databus_access_key_id'] or
          raise Fluent::ConfigError, "Credential config 'databus_access_key_id' not found."
      new_secret_access_key = section['databus_access_secret_key'] or
          raise Fluent::ConfigError, "Credential config 'databus_access_secret_key' not found."
      new_key_algo = section['databus_access_secret_key_algo'] or
          raise Fluent::ConfigError, "Credential config 'databus_access_secret_key_algo' not found."

      @load_cred_lock.with_write_lock do
        case new_key_algo.upcase
        when ED25519_ALGORITHM
          @request_signer = method(:sign_request_ed25519)
        when RSA_ALGORITHM
          @request_signer = method(:sign_request_rsa)
        else
          raise Fluent::ConfigError, "Invalid config 'databus_access_secret_key_algo', must be '#{ED25519_ALGORITHM}' or '#{RSA_ALGORITHM}'"
        end

        @databus_access_key = new_access_key
        # Config needs to escape \n and put whole key value into single line.
        # Here recover the key to correct RSA format.
        @databus_secret_access_key = new_secret_access_key.gsub("\\n", "\n")
      end
    end

    def config_api
      endpoint_uri = URI.parse(@endpoint)
      api_config = Dbus::Configuration.new
      api_config.scheme = endpoint_uri.scheme
      api_config.host = endpoint_uri.host
      api_config.debugging = @debug
      api_config.timeout=@http_timeout_seconds
      api_config.verify_ssl= @verify_ssl
      api_config.ssl_ca_cert=get_ssl_ca_cert()
      if @region.nil? || @region.empty? || @region == "us-west-1"
        log.info "Default environment region will be used for dbus api (no base path applied: /dbus)."
      else
        log.info "Found non-default environment region for dbus: #{@region}, the following base path will be used: /api/v1/dbus"
        api_config.base_path="/api/v1"
      end
      @api_config = api_config
    end

    def start
      super
      log.info "starting databus plugin"
      @api_client = @api_client_class.new(
          config = @api_config,
          request_signer = method(:request_signer_reference),
          proxy_config = @proxy_config, @debug_record_metadata, log)
      @api_instance = Dbus::DefaultApi.new(@api_client)
    end

    # We intend to put logs to dbus as mini batches in order,
    # thus implement write() method which serves sync buffered output.
    # For more details please refer to
    # https://docs.fluentd.org/plugin-development/api-plugin-output
    def write(chunk)
      raw_data = ''
      extra_headers={}
      chunk.each do |emit_time, record|
        if @event_message_field.nil? || @event_message_field.empty?
          raw_data << "#{Yajl.dump(record)}\n"
        elsif record.key?(@event_message_field)
          line = record[@event_message_field].chomp
          raw_data << "#{line}\n"
          unless @extra_headers_field.nil?
            if record.key?(@extra_headers_field)
              extra_headers=record[@extra_headers_field]
            end
          end
        end
      end

      raw_data = raw_data.chomp
      unless raw_data.empty?
        put_dbus_record(chunk.metadata.tag, raw_data, extra_headers)
      end
    end

    def put_dbus_record(tag, raw_data, extra_headers)
      headers = @headers.map {|k, v| Dbus::Header.new({:'name' => k, :'value' => v})}
      host_header = Dbus::Header.new({:'name' => 'host', :'value' => @hostname})
      headers += [host_header]
      # Tag is only available when using <buffer tag> in <match>.
      unless tag.nil?
        tag_header = Dbus::Header.new({:'name' => 'tag', :'value' => tag})
        headers += [tag_header]
      end

      unless extra_headers.empty?
        extra_headers.each { |k, v| 
          extra_header=Dbus::Header.new({:'name' => k, :'value' => v}) 
          headers+=[extra_header]
        }
      end

      raw_data_size = raw_data.bytesize
      put_record_attr = {
          :'headers' => headers,
          :'streamName' => @stream_name,
          :'partitionKey' => @partition_key,
          :'payloadSize' => raw_data_size,
      }
      if raw_data_size <= INLINE_RECORD_THRESHOLD
        put_record_attr.update({:'payload' => Base64.strict_encode64(raw_data)})
      end

      record = Dbus::Record.new(put_record_attr)
      put_record_request = Dbus::PutRecordRequest.new(
          {
              :'record' => record
          })

      put_record_response = @api_instance.put_record(put_record_request)
      if put_record_response.record.status == 'PENDING'
        log_record_metadata(put_record_response.record, headers)
        put_s3_payload(put_record_response.record.upload_url, raw_data)
      elsif put_record_response.record.status == 'SENT'
        log_record_metadata(put_record_response.record, headers)
      end
    end

    def log_record_metadata(record, headers)
      if @debug_record_metadata
        headers_map={}
        headers.each {|dbus_header| headers_map[dbus_header.name]=dbus_header.value}
        log.info("Record processed with id: #{record.record_id} [stream name: #{@stream_name}, state: #{record.status}, headers=#{headers_map}]")
      end
    end

    def put_s3_payload(presigned_url, raw_data)
      req_opts = {:method => :put, :'body' => raw_data}
      add_proxy_options(req_opts, presigned_url)
      s3_put_request = Typhoeus::Request.new(presigned_url, req_opts)
      s3_put_response = s3_put_request.run
      err_msg = "Failed to put S3 payload. #{s3_put_response.response_code}: #{s3_put_response.response_body}"

      unless s3_put_response.success?
        # Do not retry on 4xx error.
        if 400 <= s3_put_response.response_code < 500
          raise Fluent::UnrecoverableError(err_msg)
        else
          raise ApiError.new(err_msg)
        end
      end
    end

    def request_signer_reference(http_method, content_type, path)
      @load_cred_lock.with_read_lock do
        @request_signer.call(http_method, content_type, path)
      end
    end

    def sign_request_ed25519(http_method, content_type, path)
      auth_date = @now.call
      bytes_to_sign =
          "#{http_method.upcase}\n#{content_type}\n#{auth_date}\n#{path}\n#{ED25519_AUTH_METHOD}"
      if @debug_request_signer
        log.info("Message to sign: #{bytes_to_sign}")
      end
      seed = Base64.decode64(@databus_secret_access_key)
      signing_key = Ed25519::SigningKey.new(seed)
      signature = signing_key.sign(bytes_to_sign)
      auth_info = JSON[
          {
              :'access_key_id' => @databus_access_key,
              :'auth_method' => ED25519_AUTH_METHOD
          }]

      build_auth_headers(auth_date, auth_info, signature)
    end

    def sign_request_rsa(http_method, content_type, path)
      auth_date = @now.call
      bytes_to_sign =
          "#{http_method.upcase}\n#{content_type}\n#{auth_date}\n#{path}\n#{RSA_AUTH_METHOD}"
      if @debug_request_signer
        log.info("Message to sign: #{bytes_to_sign}")
      end
      key = OpenSSL::PKey::RSA.new(@databus_secret_access_key)
      signature = key.sign(OpenSSL::Digest::SHA256.new, bytes_to_sign)
      auth_info = JSON[
          {
              :'access_key_id' => @databus_access_key,
              :'auth_method' => RSA_AUTH_METHOD
          }]
      build_auth_headers(auth_date, auth_info, signature)
    end

    def build_auth_headers(auth_date, auth_info, signature)
      {
          :'x-altus-date' => auth_date,
          :'x-altus-auth' =>
              "#{Base64.urlsafe_encode64(auth_info)}.#{Base64.urlsafe_encode64(signature)}"
      }
    end

    def get_ssl_ca_cert
      cert=nil
      if @ssl_ca_cert.nil? || @ssl_ca_cert.empty?
        certs_to_find=['/etc/ssl/certs/ca-certificates.crt', '/etc/pki/tls/certs/ca-bundle.crt', '/usr/share/ssl/certs/ca-bundle.crt', '/usr/local/share/certs/ca-root-nss.crt', '/etc/ssl/cert.pem']
        for cert in certs_to_find do
            if File.exist?(cert)
                log.info "Found SSL CA cert file: #{cert}"
                return cert
            end
        end
      else
        if File.exist?(@ssl_ca_cert)
          log.info "Found SSL CA cert file: #{@ssl_ca_cert}"
        else
          raise Fluent::ConfigError, "Cannot found SSL CA certificate file #{@ssl_ca_cert}"
        end
        cert=@ssl_ca_cert
      end
      return cert
    end

    def host_in_no_proxy(no_proxy, host)
      unless no_proxy.nil? || no_proxy.empty? || host.nil?
        log.info "no_proxy is filled with #{no_proxy}, checking for presence of #{host}"
        # Logic for splitting is sourced from
        # https://github.com/ruby/ruby/blob/0ed71b37fa9af134fdd5a7fd1cebd171eba83541/lib/uri/generic.rb#L1545-L1554
        host.scan(/([^:,\s]+)(?::(\d+))?/) {|parsed_host, parsed_host_port|
          no_proxy_array = no_proxy.split(',')
          no_proxy_array.each do |item|
            item.scan(/([^:,\s]+)(?::(\d+))?/) {|item_host, item_port|
              if !item_port || parsed_host_port.to_i == item_port.to_i
                require 'ipaddr'
                item_host_ip = nil
                begin
                  item_host_ip = IPAddr.new(item_host)
                rescue IPAddr::InvalidAddressError
                  # Ignored
                end
                if item_host_ip.nil?
                  if parsed_host.include?(item_host)
                    log.info "Skipping proxy for #{host} since it is in no_proxy list"
                    return true
                  end
                else
                  begin
                    host_ip = IPSocket.getaddress(parsed_host)
                    if item_host_ip.include?(host_ip)
                      log.info "Skipping proxy for #{host} since it is in no_proxy list"
                      return true
                    end
                  rescue SocketError
                    false
                  end
                end
              end
            }
          end
        }
      end
      false
    end

    private
    def add_proxy_options(req_opts = {}, host)
      unless host_in_no_proxy(@no_proxy, host)
        unless @proxy_url.nil?
          req_opts[:proxy] = @proxy_url
          unless @proxy_username.nil? || @proxy_password.nil?
            req_opts[:proxyuserpwd] = "#{@proxy_username}:#{@proxy_password}"
          end
        end
      end
    end

  end
end
